@@ -20,6 +20,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
 #include <math.h>
 #include <stdio.h>
 #include <string.h>
@@ -32,10 +33,11 @@
 
 #include <libgnomecanvas/gnome-canvas.h>
 #include <libgnomecanvas/gnome-canvas-util.h>
+//#include <libgnomeprint/gnome-print-pango.h>
 #include "gcp-canvas-rich-text.h"
 #include "gcp-canvas-i18n.h"
-#include "gnome-print-pango.h"
 #include "gprintable.h"
+void gnome_print_pango_layout_print (GnomePrintContext *gpc, PangoLayout* pl);
 
 struct _GnomeCanvasRichTextExtPrivate {
 	GtkTextLayout *layout;
@@ -65,6 +67,7 @@ struct _GnomeCanvasRichTextExtPrivate {
 	int left_margin;
 	int right_margin;
 	int indent;
+	GtkTextDirection text_direction;
 
 	guint preblink_timeout;
 	guint blink_timeout;
@@ -75,6 +78,7 @@ struct _GnomeCanvasRichTextExtPrivate {
 	gint drag_start_y;
 
 	gboolean just_selected_element;
+	gboolean rendered;
 
 	int clicks;
 	guint click_timeout;
@@ -102,6 +106,7 @@ enum {
 	PROP_LEFT_MARGIN,
 	PROP_RIGHT_MARGIN,
 	PROP_INDENT,
+	PROP_TEXT_DIRECTION,
 };
 
 enum {
@@ -145,10 +150,12 @@ static void gnome_canvas_rich_text_ext_move_cursor(GnomeCanvasRichTextExt *text,
 					       gint count,
 					       gboolean extend_selection);
 static void   gnome_canvas_rich_text_ext_print       (GPrintable *gprintable, GnomePrintContext *pc);
-
+static void gnome_canvas_rich_text_ext_export_svg   (GPrintable *gprintable, xmlDocPtr doc, xmlNodePtr node);
+static void gnome_canvas_rich_text_set_cursor_direction (GnomeCanvasRichTextExt *text);
 
 static GtkTextBuffer *get_buffer(GnomeCanvasRichTextExt *text);
 static gint blink_cb(gpointer data);
+static void gnome_canvas_rich_text_ext_ensure_rendered(GnomeCanvasRichTextExt *text);
 
 #define PREBLINK_TIME 300
 #define CURSOR_ON_TIME 800
@@ -158,6 +165,7 @@ static void
 gnome_canvas_rich_text_ext_print_init (GPrintableIface *iface)
 {
 	iface->print = gnome_canvas_rich_text_ext_print;
+	iface->export_svg = gnome_canvas_rich_text_ext_export_svg;
 }
 
 GType
@@ -389,6 +397,15 @@ gnome_canvas_rich_text_ext_class_init(GnomeCanvasRichTextExtClass *klass)
 				  G_MININT, G_MAXINT,
 				  0,
 				  G_PARAM_READWRITE));
+	g_object_class_install_property (
+		gobject_class,
+		PROP_TEXT_DIRECTION,
+		g_param_spec_int ("text_direction",
+				  _("Text direction"),
+				  _("Text direction (rtl, ltr or default)"),
+				  GTK_TEXT_DIR_NONE, GTK_TEXT_DIR_RTL,
+				  GTK_TEXT_DIR_NONE,
+				  G_PARAM_READWRITE));
 
 	/* Signals */
 	signals[TAG_CHANGED] = g_signal_new(
@@ -433,12 +450,14 @@ gnome_canvas_rich_text_ext_init(GnomeCanvasRichTextExt *text)
 	text->_priv->justification = GTK_JUSTIFY_LEFT;
 	text->_priv->direction = gtk_widget_get_default_direction();
 	text->_priv->anchor = GTK_ANCHOR_NW;
+	text->_priv->text_direction = GTK_TEXT_DIR_NONE;
 	
 	text->_priv->blink_timeout = 0;
 	text->_priv->preblink_timeout = 0;
 
 	text->_priv->clicks = 0;
 	text->_priv->click_timeout = 0;
+	text->_priv->rendered = FALSE;
 } /* gnome_canvas_rich_text_ext_init */
 
 static void
@@ -603,6 +622,11 @@ gnome_canvas_rich_text_ext_set_property (GObject *object, guint property_id,
 			gtk_text_layout_default_style_changed(text->_priv->layout);
 		}
 		break;
+	case PROP_TEXT_DIRECTION:
+		text->_priv->text_direction = (GtkTextDirection) g_value_get_int (value);
+		if (text->_priv->layout)
+			gnome_canvas_rich_text_set_cursor_direction (text);
+		break;
 		       
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -670,6 +694,8 @@ gnome_canvas_rich_text_ext_get_property (GObject *object, guint property_id,
 	case PROP_INDENT:
 		g_value_set_int (value, text->_priv->indent);
 		break;
+	case PROP_TEXT_DIRECTION:
+		g_value_set_int (value, (int) text->_priv->text_direction);
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
 		break;
@@ -691,6 +717,9 @@ gnome_canvas_rich_text_ext_unrealize(GnomeCanvasItem *item)
 {
 	GnomeCanvasRichTextExt *text = GNOME_CANVAS_RICH_TEXT_EXT(item);
 
+	/*remove idle calls to request_update */
+	while (	g_idle_remove_by_data (item));
+
 	gnome_canvas_rich_text_ext_destroy_layout(text);
 
 	(* GNOME_CANVAS_ITEM_CLASS(parent_class)->unrealize)(item);
@@ -1670,6 +1699,8 @@ gnome_canvas_rich_text_ext_stop_cursor_blink(GnomeCanvasRichTextExt *text)
 static gboolean
 request_update(gpointer data)
 {
+	if (!GNOME_IS_CANVAS_RICH_TEXT_EXT (data))
+		return;
 	GnomeCanvasRichTextExt *text = GNOME_CANVAS_RICH_TEXT_EXT(data);
 
 	gnome_canvas_item_request_update(GNOME_CANVAS_ITEM(text));
@@ -1835,6 +1866,40 @@ gnome_canvas_rich_text_ext_set_attributes_from_style(GnomeCanvasRichTextExt *tex
 
 } /* gnome_canvas_rich_text_ext_set_attributes_from_style */
 
+static void
+gnome_canvas_rich_text_set_cursor_direction (GnomeCanvasRichTextExt *text)
+{
+	GtkTextDirection new_cursor_dir;
+	GtkTextDirection new_keyboard_dir;
+	gboolean split_cursor;
+	GtkWidget *canvas;
+	GtkSettings *settings; 
+	GdkKeymap *keymap;
+
+	canvas = GTK_WIDGET(GNOME_CANVAS_ITEM(text)->canvas);
+	settings = gtk_widget_get_settings (canvas);
+	g_object_get (settings,
+		"gtk-split-cursor", &split_cursor,
+		NULL);
+
+	if (text->_priv->text_direction == GTK_TEXT_DIR_NONE) {
+		keymap = gdk_keymap_get_for_display (gtk_widget_get_display (canvas));
+		if (gdk_keymap_get_direction (keymap) == PANGO_DIRECTION_LTR)
+			new_keyboard_dir = GTK_TEXT_DIR_LTR;
+		else
+			new_keyboard_dir  = GTK_TEXT_DIR_RTL;
+	} else
+		new_keyboard_dir = text->_priv->text_direction;
+
+	if (split_cursor)
+		new_cursor_dir = GTK_TEXT_DIR_NONE;
+	else
+		new_cursor_dir = new_keyboard_dir;
+
+	gtk_text_layout_set_cursor_direction (text->_priv->layout, new_cursor_dir);
+	gtk_text_layout_set_keyboard_direction (text->_priv->layout, new_keyboard_dir);
+}
+
 static void
 gnome_canvas_rich_text_ext_ensure_layout(GnomeCanvasRichTextExt *text)
 {
@@ -1866,6 +1931,7 @@ gnome_canvas_rich_text_ext_ensure_layout(GnomeCanvasRichTextExt *text)
 		pango_context_set_base_dir(ltr_context, PANGO_DIRECTION_LTR);
 		rtl_context = gtk_widget_create_pango_context(canvas);
 		pango_context_set_base_dir(rtl_context, PANGO_DIRECTION_RTL);
+		gnome_canvas_rich_text_set_cursor_direction (text);
 
 		gtk_text_layout_set_contexts(
 			text->_priv->layout, ltr_context, rtl_context);
@@ -2208,6 +2274,7 @@ gnome_canvas_rich_text_ext_render(GnomeCanvasItem *item, GnomeCanvasBuf *buf)
 	g_object_unref (gc);
 	g_object_unref (pb);
 	g_object_unref (pxb);
+	text->_priv->rendered = TRUE;
 } /* gnome_canvas_rich_text_ext_render */
 
 #if 0
@@ -2275,28 +2342,19 @@ PangoLayout* gnome_canvas_rich_text_ext_get_pango_layout (GnomeCanvasRichTextExt
 static void 
 gnome_canvas_rich_text_ext_print       (GPrintable *gprintable, GnomePrintContext *pc)
 {
-	gint i, top, bottom, firsty;
+	gint top, bottom, firsty;
 	GSList *lines;
 	GtkTextLine *text_line;
-	PangoLayoutIter *iter;
-	PangoGlyphItem *item;
-	PangoRectangle logical_rect;
 	GtkTextLineDisplay *display;
 	GtkTextIter text_iter;
 	GdkRectangle rect;
 	GnomeCanvasRichTextExt *text;
-	GnomeFont *font;
-	GnomeFontFace *face;
-	GnomeGlyphList *glyph_list;
-	PangoFontDescription *desc;
-	GSList *extra_attrs_list;
-	gboolean foreground_set = FALSE;
-	guint16 foreground_red = 0, foreground_green = 0, foreground_blue = 0;
-	gboolean background_set = FALSE;
-	guint16 background_red = 0, background_green = 0, background_blue = 0;
 	g_return_if_fail (GNOME_IS_CANVAS_RICH_TEXT_EXT (gprintable));
 	text = GNOME_CANVAS_RICH_TEXT_EXT (gprintable);
-	g_return_if_fail (gtk_text_buffer_get_char_count (text->_priv->buffer) > 0);
+	if (gtk_text_buffer_get_char_count (text->_priv->buffer) == 0)
+		return;
+	if (!text->_priv->rendered)
+		gnome_canvas_rich_text_ext_ensure_rendered (text);
 	gtk_text_layout_get_size (text->_priv->layout, NULL, &bottom);
 	top = 0;
 	if (bottom <= top) return;
@@ -2310,10 +2368,70 @@ gnome_canvas_rich_text_ext_print       (GPrintable *gprintable, GnomePrintContex
 		top = rect.y;
 		gnome_print_gsave (pc);
 		gnome_print_translate (pc, text->_priv->x, text->_priv->y + top);
-		gnome_print_pango_layout_print (pc, display->layout);
+#ifdef GP_HAS_PANGO
+		pango_layout_print (pc, display->layout);
+#else	
+		gpc_print_pango_layout_print (pc, display->layout);
+#endif
 		gnome_print_grestore (pc);
-		gnome_canvas_rich_text_ext_get_iter_location (text, &text_iter, &rect);
 		lines = lines->next;
 		firsty++;
 	}
 }
+
+extern void pango_layout_to_svg (PangoLayout* layout, xmlDocPtr doc, xmlNodePtr node, double x, double y);
+
+static void
+gnome_canvas_rich_text_ext_export_svg (GPrintable *printable, xmlDocPtr doc, xmlNodePtr node)
+{
+	gint top, bottom, firsty;
+	GSList *lines;
+	GtkTextLine *text_line;
+	GtkTextLineDisplay *display;
+	GtkTextIter text_iter;
+	GdkRectangle rect;
+	GnomeCanvasRichTextExt *text;
+	xmlNodePtr child;
+	char *buf;
+	g_return_if_fail (GNOME_IS_CANVAS_RICH_TEXT_EXT (printable));
+	text = GNOME_CANVAS_RICH_TEXT_EXT (printable);
+	if (!text->_priv->rendered)
+		gnome_canvas_rich_text_ext_ensure_rendered (text);
+	if (gtk_text_buffer_get_char_count (text->_priv->buffer) == 0)
+		return;
+	gtk_text_layout_get_size (text->_priv->layout, NULL, &bottom);
+	top = 0;
+	if (bottom <= top) return;
+	lines = gtk_text_layout_get_lines (text->_priv->layout, top, bottom, &firsty);
+	firsty = 0;
+	while (lines) {
+		text_line = lines->data;
+		display = gtk_text_layout_get_line_display (text->_priv->layout, text_line, TRUE);
+		gtk_text_buffer_get_iter_at_line (text->_priv->buffer, &text_iter, firsty);
+		gnome_canvas_rich_text_ext_get_iter_location (text, &text_iter, &rect);
+		top = rect.y;
+		pango_layout_to_svg (display->layout, doc, node, text->_priv->x, text->_priv->y + top);
+		lines = lines->next;
+		firsty++;
+	}
+}
+
+static void
+gnome_canvas_rich_text_ext_ensure_rendered (GnomeCanvasRichTextExt *text)
+{
+	double x1, y1, x2, y2;
+	get_bounds (text, &x1, &y1, &x2, &y2);
+	int w = (int) (ceil (x2) - floor (x1)) + 1, h = (int) (ceil (y2) - floor (y1)) + 1;
+	GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, w, h);
+	GnomeCanvasBuf buf;
+	buf.buf = gdk_pixbuf_get_pixels (pixbuf);
+	buf.rect.x0 = (int) floor (x1);
+	buf.rect.x1 = (int) ceil (x2);
+	buf.rect.y0 = (int) floor (y1);
+	buf.rect.y1 = (int) ceil (y2);
+	buf.buf_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
+	buf.bg_color = 0xffffff;
+	buf.is_buf = 1;
+	gnome_canvas_rich_text_ext_render (GNOME_CANVAS_ITEM (text), &buf);
+	g_object_unref (pixbuf);
+}
