@@ -371,6 +371,11 @@ static bool do_save_as_image(GtkWidget* widget, GtkFileSelection *fsel)
 		if (result)
 			g_free (result);
 	}
+	else if (!strcmp (mime_type, "image/svg+xml")) {
+		pActiveDoc->GetView ()->ExportImage (filename, "svg");
+		gtk_widget_destroy (GTK_WIDGET (fsel));
+		return false;
+	}
 	else {
 		GtkWidget* message = gtk_message_dialog_new (GTK_WINDOW (fsel), GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
 															_("Sorry, format not supported!"));
@@ -617,6 +622,8 @@ int main(int argc, char *argv[])
 		SubscriptBtn = glade_xml_get_widget(xml, "subscript");
 		SuperscriptBtn = glade_xml_get_widget(xml, "superscript");
 		FontBtn = glade_xml_get_widget(xml, "font");
+		gtk_toolbar_set_show_arrow(GTK_TOOLBAR (TextToolBar), false);
+		gtk_widget_show_all (TextToolBar);
 		gnome_app_add_toolbar(app, GTK_TOOLBAR(widget), "texttools", BONOBO_DOCK_ITEM_BEH_NORMAL, BONOBO_DOCK_TOP, 2, 0, 0);
 		gtk_widget_set_sensitive(widget, false);
 		g_object_unref(G_OBJECT(widget));
